@@ -10,10 +10,10 @@ $!  = ''            unless defined $!;
 $;  = chr(28)       unless defined $;;
 $?  = 0             unless defined $?;
 # $[  = 0             unless defined $[;    # "assignment to $[ is deprecated"
-$]  = '5.020000'    unless $];      # $] is defined(), but ${"main::]"} is not
-$^V = bless( { 'original' => 'v5.20.0',
+$]  = '5.022000'    unless $];      # $] is defined(), but ${"main::]"} is not
+$^V = bless( { 'original' => 'v5.22.0',
                'qv'       => 1,
-               'version'  => [ 5, 20, 0 ]
+               'version'  => [ 5, 22, 0 ]
              }, 'version' )
                     unless defined $^V;
 
@@ -23,24 +23,31 @@ our $STRICT       = 0;
 our $WARNINGS     = 0;
 our $UTF8         = 0;
 our $BYTES        = 0;
-our $CALLER       = [];
+our @CALLER       = ();
 our %DATA_SECTION = ();   # contents of the __DATA__ sections per package
 our $PKG_NAME     = '';   # current package being compiled
 our $LINE_NUMBER  = 0;    # current line number being compiled
 our $FILE_NAME    = '';   # current file name being compiled
 
 # information about the current compilation process
-our $GLOBAL       = {};
-our $BASE_SCOPE   = Perlito5::Grammar::Scope->new_base_scope();
-our $SCOPE        = $BASE_SCOPE;    # information about the current block being compiled
-our $SCOPE_DEPTH  = 0;
+our $GLOBAL          = {};
+our $BASE_SCOPE      = Perlito5::Grammar::Scope->new_base_scope();
+our $CLOSURE_SCOPE   = $BASE_SCOPE;    # variables that are in scope in the current closure being compiled
+our $SCOPE           = $BASE_SCOPE;    # variables that are in scope in the current block being compiled
+our $SCOPE_DEPTH     = 0;
 our @SCOPE_STMT      = ();
 our @END_BLOCK       = ();    # END block LIFO - array of subs
 our @INIT_BLOCK      = ();    # INIT block FIFO - array of subs
 our @CHECK_BLOCK     = ();    # CHECK block LIFO - array of subs
 our @UNITCHECK_BLOCK = ();    # UNITCHECK block LIFO - array of subs
+our %BEGIN_SCRATCHPAD = (); # list of "my" variables captured in BEGIN blocks
 our $PROTO           = {};
 
+# the Perl-to-Java compiler uses this syntax for "annotations":
+#   package Put { import => 'java.Put' };
+# annotations are stored as namespace + AST
+our @ANNOTATION;
+
 sub set_global_phase {
     my $phase = shift;
     local $@;
@@ -289,7 +296,7 @@ our $CORE_PROTO = {
           'CORE::length' => '_',
           'CORE::state' => undef,
           'CORE::die' => '@',
-          'CORE::delete' => undef,
+          'CORE::delete' => '$',    # original 'undef'
           'CORE::getservent' => '',
           'CORE::getservbyname' => '$$',
           'CORE::setpriority' => '$$$',
