@@ -46,9 +46,13 @@ sub term_bareword {
     }
 
     my $name = Perlito5::Match::flat($m_name);
+    if ($name eq '__PACKAGE__' && $namespace eq '') {
+        $m_name->{capture} = [ 'term', Perlito5::AST::Apply->new( code => $name, namespace => '', arguments => [], bareword => 1 ) ];
+        return $m_name;
+    }
     $p = $m_name->{to};
 
-    if ( substr( $str, $p, 2) eq '::' ) {
+    if ( $str->[$p] eq ':' && $str->[$p+1] eq ':' ) {
         # ::X::y::
         $m_name->{to} = $p + 2;
         $m_name->{capture} = [ 'term', 
@@ -73,6 +77,20 @@ sub term_bareword {
         $p = $m->{to};
     }
 
+    if ( $str->[$p] eq '!' && ( $str->[$p+1] eq '=' || $str->[$p+1] eq '~' ) ) {
+        # != or !~
+        # "X::y != ..."  subroutine call
+        $m_name->{capture} = [ 'term', 
+                    Perlito5::AST::Apply->new(
+                        code      => $name,
+                        namespace => $namespace,
+                        arguments => [],
+                        bareword  => 1
+                    )
+                ];
+        return $m_name;
+    }
+
     # check for indirect-object
     my $invocant;
     my $is_subroutine_name;
@@ -99,25 +117,27 @@ sub term_bareword {
         # this can be an indirect-object if the next term is a bareword ending with '::'
 
         $invocant = Perlito5::Grammar::full_ident( $str, $p );
-        my $package = Perlito5::Match::flat($invocant);
-        if ( $package ) {
-            $invocant->{capture} = Perlito5::AST::Var->new(
-                                     sigil => '::',
-                                     name  => '',
-                                     namespace => $package,
-                                 );
-            if ( substr( $str, $invocant->{to}, 2) eq '::' ) {
-                # ::X::y::
-                $invocant->{to} = $invocant->{to} + 2;
-            }
-            else {
-                # is this a known package name?
-                if ( ! $Perlito5::PACKAGES->{ $package } ) {
-                    # not a known package name
-                    $invocant = undef;
+        if ($invocant) {
+            my $package = Perlito5::Match::flat($invocant);
+            if ( $package ) {
+                $invocant->{capture} = Perlito5::AST::Var->new(
+                                         sigil => '::',
+                                         name  => '',
+                                         namespace => $package,
+                                     );
+                if ( $str->[$invocant->{to}] eq ':' && $str->[$invocant->{to}+1] eq ':' ) {
+                    # ::X::y::
+                    $invocant->{to} = $invocant->{to} + 2;
                 }
                 else {
-                    # valid package name - this is indirect-object
+                    # is this a known package name?
+                    if ( ! $Perlito5::PACKAGES->{ $package } ) {
+                        # not a known package name
+                        $invocant = undef;
+                    }
+                    else {
+                        # valid package name - this is indirect-object
+                    }
                 }
             }
         }
@@ -135,9 +155,10 @@ sub term_bareword {
         my $arg = [];
         $m = Perlito5::Grammar::Space::ws( $str, $p );
         $p = $m->{to} if $m;
-        if ( substr($str, $p, 2) eq '->' ) {
+        if ( $str->[$p] eq '-' && $str->[$p+1] eq '>' ) {
+            # ->
         }
-        elsif ( substr($str, $p, 1) eq '(' ) {
+        elsif ( $str->[$p] eq '(' ) {
             my $m = Perlito5::Grammar::Expression::term_paren( $str, $p );
             if ( $m ) {
                 $arg = $m->{capture}[2];
@@ -164,7 +185,8 @@ sub term_bareword {
         return $m_name;
     }
 
-    if ( substr( $str, $p, 2 ) eq '=>' ) {
+    if ( $str->[$p] eq '=' && $str->[$p+1] eq '>' ) {
+        # =>
         # autoquote bareword
         $m_name->{capture} = [ 'term', 
                     Perlito5::AST::Apply->new(
@@ -177,7 +199,8 @@ sub term_bareword {
         $m_name->{to} = $p;
         return $m_name;
     }
-    if ( substr( $str, $p, 2 ) eq '->' ) {
+    if ( $str->[$p] eq '-' && $str->[$p+1] eq '>' ) {
+        # ->
         if ( $is_subroutine_name ) {
             # call()->method call
             $m_name->{capture} = [ 'term', 
@@ -289,7 +312,7 @@ sub term_bareword {
         if ($sig_part eq '&') {
             $m = Perlito5::Grammar::Space::ws( $str, $p );
             $p = $m->{to} if $m;
-            if ( substr($str, $p, 1) ne '(' ) {
+            if ( $str->[$p] ne '(' ) {
                 $sig = substr($sig, 1);
                 $m = Perlito5::Grammar::Bareword::prototype_is_ampersand( $str, $p );
                 $capture = $m->{capture} if $m;
@@ -303,7 +326,7 @@ sub term_bareword {
             }
         }
 
-        if ( substr($sig, 0, 1) eq ';' && substr($str, $p, 2) eq '//' ) {
+        if ( substr($sig, 0, 1) eq ';' && $str->[$p] eq '/' && $str->[$p+1] eq '/' ) {
             # argument is optional - shift(), pop() followed by //
             #
             # special case - see test t5/01-perlito/25-syntax-defined-or.t
@@ -323,15 +346,15 @@ sub term_bareword {
 
         if ( $sig eq '' ) {
             # empty sig - we allow (), but only if it is empty
-            if ( substr($str, $p, 1) eq '(' ) {
+            if ( $str->[$p] eq '(' ) {
                 $p++;
                 $has_paren = 1;
                 my $m = Perlito5::Grammar::Space::ws( $str, $p );
                 if ($m) {
                     $p = $m->{to}
                 }
-                if ( substr($str, $p, 1) ne ')' ) {
-                    Perlito5::Compiler::error( "syntax error near ", substr( $str, $pos, 10 ));
+                if ( $str->[$p] ne ')' ) {
+                    Perlito5::Compiler::error( "syntax error near ", join("", @{$str}[ $pos .. $pos + 10 ] ));
                 }
                 $p++;
             }
@@ -428,7 +451,7 @@ sub term_bareword {
         if ( $sig eq '_' || $sig eq '$' || $sig eq '+' || $sig eq ';$' ) {
             my $m;
             my $arg;
-            if ( substr($str, $p, 1) eq '(' ) {
+            if ( $str->[$p] eq '(' ) {
                 $m = Perlito5::Grammar::Expression::term_paren( $str, $p );
                 if ( !$m ) { return $m };
                 $p = $m->{to};
@@ -476,14 +499,17 @@ sub term_bareword {
                     );
             if ($name eq 'eval' && !$namespace) {
                 # add scope information to eval-string
-                $ast->{_scope} = Perlito5::Grammar::Scope::get_snapshot();
+                # - here we add the variables declared so far in the current closure
+                # - other variables captured by the current closure need to be added when the closure finishes compiling
+
+                $ast->{_scope} = Perlito5::Grammar::Scope::get_snapshot( $Perlito5::CLOSURE_SCOPE );
             }
             $m->{capture} = [ 'term', $ast ];
             return $m;
         }
 
-        if ( $sig eq ';@' ) {
-            if ( substr($str, $p, 1) eq '(' ) {
+        if ( $sig eq ';@' || $sig eq '@' ) {
+            if ( $str->[$p] eq '(' ) {
                 $m = Perlito5::Grammar::Expression::term_paren( $str, $p );
                 $has_paren = 1;
                 my $arg = $m->{capture}[2];
@@ -526,7 +552,7 @@ sub term_bareword {
 
     # maybe it's a subroutine call
 
-    if ( substr($str, $p, 1) eq '(' ) {
+    if ( $str->[$p] eq '(' ) {
         $m = Perlito5::Grammar::Expression::term_paren( $str, $p );
         if ( !$m ) { return $m };
         my $arg = $m->{capture}[2];
@@ -643,6 +669,26 @@ sub term_bareword {
 
     # it's just a bareword - we will disambiguate later
 
+    # TODO - forbid barewords that are not "GLOB" - "close FILE" is ok
+    #
+    # if ($Perlito5::STRICT) {
+    #     if (
+    #         !(
+    #             exists( $Perlito5::PROTO->{$effective_name} )
+    #             || (
+    #                 ( !$namespace || $namespace eq 'CORE' )
+    #                 && exists $Perlito5::CORE_PROTO->{"CORE::$name"}    # subroutine comes from CORE
+    #             )
+    #         )
+    #       )
+    #     {
+    #         # subroutine was not predeclared
+    #         # print STDERR "effective_name [$effective_name]\n";
+    #         Perlito5::Compiler::error 'Bareword "' . ( $namespace ? "${namespace}::" : "" ) . $name . '" not allowed while "strict subs" in use';
+
+    #     }
+    # }
+
     $m_name->{capture} = [ 'postfix_or_term', 'funcall_no_params',
             $namespace,
             $name
